-- Név: sor típus (adott hosszal, mely a típusban rögzített)
-- Ez a modul egy sort valósít meg funkcionálisan (ez egyfajta konvencionális,
-- szokásos implementáció, leírása:
-- www.cs.cmu.edu/~rwh/theses/okasaki.pdf‎
-- Chris Okasaki: Purely functional data structures
module queue

-- Az s : Nat a sor indexe, ez adja meg a belerakott elemek számát (ez nőhet,
-- ekkor egy más típusú sort kapunk)
-- Az a : Type az elemek típusa
abstract
data Queue : (s : Nat) -> (a : Type) -> Type where
    queue : (Vect n a) -> (Vect m a) -> (Queue (n + m) a)

-- Smart contructor a fenti cikk szerint, mely tartja az invariánst
private
queue' : (Vect n a) -> (Vect m a) -> (Queue (n + m) a)
queue' [] y ?= queue (reverse y) []
queue' x y  ?= queue x y

-- Üres sor
abstract
emptyQueue : (Queue Z a)
emptyQueue = queue' [] []

-- Üres-e a sor
abstract
isEmptyQueue : (Queue n a) -> Bool
isEmptyQueue (queue' [] []) = True
isEmptyQueue _ = False

-- Push művelet (a cikkben is így hívják, ezért hívom én is így)
-- Láthatjuk, hogy a más típusút ad vissza, azaz eggyel nagyobb méretűt
abstract
push : (Queue n a) -> a -> (Queue (S n) a)
push (queue x y) elem ?= (queue' x (elem :: y))

-- Pop művelet (a cikkben is így hívják, ezért hívom én is így)
-- Láthatjuk, hogy a más típusút ad vissza, azaz eggyel kisebb méretűt
abstract
pop : (Queue (S n) a) -> (Queue n a)
pop (queue (x :: xs) y) = (queue' xs y)

-- A legelső elem lekérdezése
abstract
front : (Queue n a) -> a
front (queue (x :: _) _) = x

-- A Show típusosztály példánya (kiírás stringként)
instance (Show a) => Show (Queue n a) where
    show (queue a b) = "->" ++ (show (a ++ (reverse b))) ++ "#"

-- Főprogram, példa
-- 1, létrehoz egy üres sort (Nat elemekkel, az üres sor elemeinek típusa is fontos)
-- 2, Belerakja az alábbi elemeket: 2, 3, 4, 10
-- 3, Kiírja a képernyőre a sor tartalmát
-- 4, Kivesz egy elemet és ismét kiírja
-- 5, Az előzőnek kiírja az első elemét
-- Meghívás:
-- :l queue.idr
-- :x main
public
main : IO ()
main = do let q = (the (Queue Z Nat) emptyQueue)
          let q2 = (push (push (push (push q 2) 3) 4) 10)
          putStrLn $ show q2
          putStrLn $ show $ pop q2
          putStrLn $ show $ front $ pop q2


-- Bizonyítások
-- Ezek a program írása közben jelentek meg tételekként, be kellett bizonyítani
-- őket (úgynevezett lyukként jelentek meg, az interaktív bizonyítórendszerrel
-- történt a bizonyításuk)
----------------------------
---------- Proofs ----------

-- A típusokra tekintsünk úgy, mintha implikációk lennének, például: A -> B
-- Bizonyítandó a jobb oldal
-- intros - vedd fel feltevésként, amit csak lehet
-- intro  - végy fel egyet a következtetés bal oldaláról feltevésként, ha lehet
-- rewrite - írd át adott szabály szerint
-- compute - egyszerűsítsd a kifejezést
-- trivial - vedd észre, hogy triviális az állítás

-- A bizonyítások menete analóg az idris-tutorial 9.1-gyel
--  Bizonyítandó ekvivalencia: (Queue (n + (S m)) a) ~ Queue (S (plus n m)) a
queue.push_lemma_1 = proof
  intro
  intro
  intro
  intro
  intro
  intro
  compute
  rewrite sym (plusSuccRightSucc n m)
  intro
  trivial

--  Bizonyítandó ekvivalencia: (Queue (plus n m) a) ~ Queue (plus n m) a
queue.queue'_lemma_2 = proof
  intros
  trivial

--  Bizonyítandó ekvivalencia: (Queue (plus m Z) a) ~ Queue m a
queue.queue'_lemma_1 = proof
  intros
  rewrite (plusZeroRightNeutral m)
  trivial


